#ifndef DOM_PARALLEL_HPP
#  define DOM_PARALLEL_HPP

#  include "dom.hpp"
#  include "adv.hpp"

template <class unit, typename real_t>
class dom_parallel : public dom<unit, real_t>
{
  private: int nk;
  private: dom_serial<unit, real_t> **doms;
  private: adv<unit, real_t> *fllbck, *advsch;
  
  public: dom_parallel(adv<unit, real_t> *fllbck, adv<unit, real_t> *advsch, int nx, int nk)
    : nk(nk), fllbck(fllbck), advsch(advsch)
  {
    // subdomain length
    int nxs = nx / nk;
    if (nxs != ((1.*nx) / (1.*nk))) error_macro("nx/nk must be an integer value (" << nx << "/" << nk << " given)")

    // domain allocation
    doms = new dom_serial<unit, real_t>*[nk];
    for (int k=0; k < nk; ++k) 
      doms[k] = new dom_serial<unit, real_t>(fllbck, advsch, k * nxs, (k+1) * nxs - 1);

    // periodic boundary over prallel domain
    for (int k=0; k < nk; ++k) 
    {   
      int l = (nk + k - 1) % nk; 
      int r = (nk + k + 1) % nk;  
      doms[k]->hook_neighbours(doms[l], doms[r]);
    }
  }

  public: ~dom_parallel()
  {
    for (int k=0; k < nk; ++k) delete doms[k];
    delete[] doms;
  }

  private: virtual void barrier() = 0;

  private: void output(int n, bool endrec = true)
  {
    for (int k=0; k < nk; ++k) doms[k]->output(n, false);
    if (endrec) cout << endl << endl;
  }

  public: void integ_loop_k(int nt, quantity<si::dimensionless, real_t> &C, int k) 
  {
    int n = 0;
    for (int t = 0; t < nt; ++t)
    {   
      adv<unit, real_t> *a;
      bool fallback = this->choose_an(&a, &n, t, advsch, fllbck);
      barrier();
      if (k == 0) output(n);
      for (int s = 1; s <= a->num_steps(); ++s)
      {   
        barrier();
        doms[k]->fill_halos(n);
        barrier();
        doms[k]->advect(a, n, C, s); 
        if (!fallback) doms[k]->cycle_arrays(n);
      } // s
    } // t
    barrier();
    if (k == 0) output(n);
  }

  // e.g. for MPI + OpenMP nesting (FIXME)
  private: quantity<unit, real_t> *rght_edge(int n) { throw; }
  private: quantity<unit, real_t> *left_edge(int n) { throw; }
};
#endif
