#ifndef MODEL_HPP
#  define MODEL_HPP

#  include "blitz-gnuplot.hpp"

#  include "adv_upstream.hpp"
#  include "adv_mpdata.hpp"
#  include "adv_leapfrog.hpp"

#  include "dom_serial.hpp"
#  ifdef _OPENMP
#    include "dom_parallel_openmp.hpp"
#  endif
#  include "dom_parallel_threads.hpp"

#  include <boost/lexical_cast.hpp>

template <typename real_t>
void model(const po::variables_map& vm) 
{
  // parameters 
  if (!vm.count("nx") || !vm.count("nt") || !vm.count("dx") || !vm.count("dt") || !vm.count("u"))
    error_macro("nx, nt, dx, dt and u options are mandatory")
  quantity<si::length, real_t> dx = boost::lexical_cast<real_t>(vm["dx"].as<string>()) * si::metres;
  quantity<si::time, real_t> dt = boost::lexical_cast<real_t>(vm["dt"].as<string>()) * si::seconds;
  quantity<si::velocity, real_t> u = boost::lexical_cast<real_t>(vm["u"].as<string>()) * si::metres / si::seconds;
  quantity<si::dimensionless, real_t> C = u * dt / dx;
  int nx = vm["nx"].as<int>();
  int nt = vm["nt"].as<int>();

  // advection sheme "allocation" (FIXME: no deallocation yet)
  if (!vm.count("adv")) error_macro("advection scheme not specified (--adv option)");
  string advscheme = vm["adv"].as<string>();
  adv<si::dimensionless, real_t> *advsch, *fllbck = NULL;
  if (advscheme == "upstream") 
    advsch = new adv_upstream<si::dimensionless, real_t>();
  else if (advscheme == "leapfrog")
  {
    advsch = new adv_leapfrog<si::dimensionless, real_t>();
    fllbck = new adv_upstream<si::dimensionless, real_t>();
  }
  else if (advscheme == "mpdata")
    advsch = new adv_mpdata<si::dimensionless, real_t>();
  else error_macro("unsupported advection scheme: " << advscheme);

  // domain set-up (FIXME: no deallocation yet)
  if (!vm.count("dom")) error_macro("domain type not specified (--dom option)");
  string domtype = vm["dom"].as<string>();
  dom<si::dimensionless, real_t> *domain;
  if (domtype == "serial")
    domain = new dom_serial<si::dimensionless, real_t>(fllbck, advsch, 0, nx - 1);
  else 
  {
    if (!vm.count("nk")) error_macro("subdomain count not specified (--nk option)");
    int nk = vm["nk"].as<int>();
#  ifdef _OPENMP
    if (domtype == "openmp")
      domain = new dom_parallel_openmp<si::dimensionless, real_t>(fllbck, advsch, nx, nk);
    else
#  endif
    if (domtype == "threads")
      domain = new dom_parallel_threads<si::dimensionless, real_t>(fllbck, advsch, nx, nk);
    else 
      error_macro("unsupported domain type: " << domtype);
  }

  // time integration loop
  domain->integ_loop(nt, C);
}
#endif
